/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          eqlt version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx


// These variables below can be directly modified by the user on the options form: 

var item_cnt= 5;       // number of stimuli in the task
  
var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 600;     // Task time in seconds if clock = 1000ms (default). 

var is_back = true;    // if (is_back==true) => user can go back to previous task/item.  

var prg_type = '0';   //  type of information in Progress Bar ('0' => Items, '1' => Time; '2' => Nothing) 
    		
var demo = false;     // if (demo == true) => Show task solutions 

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global constants and variables:

const clock = 1000;  // basic timer resolution in ms

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var counter = 0;      // timer counter. if (counter >= time) Timer will call theEnd

var is_task = false;    // task in progress or not for pageResize() event. is_task == true for all events between startTask() and theEnd() events

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var item = 0;   // current item in task     

var is_storage= false; // local storage is enabled on user browser or not

var is_move= false;  // user can move (drag and drop) or not (see events: onMouseDown, onMouseUp, mouseMove,..)

//  global arrays declarations..
 
/* stimulus => item (0..item_cnt-1) x line (0, 1) * coord (0 = x, 1 = y) 
	 (last dimension x, y: units are percents of canvas.width or height) */
var stim = [];    

// dependent variable (error) for each item in range: 0..item_cnt-1.  Unit: %
var dv = [];

// missing values for each item in range: 0..item_cnt-1
var mis = [];
  
// line1 (fixed line): slope for each item [0..item_cnt-1] dy/dx= (y2 - y1) / (x2 - x1) => const	
var k1 = [];


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


// application init [create/start Timer, check storage, show start panel..] 
function init() 
{    
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);    
  is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }				  
	setIns();		
	showStart();  
}


// application timer handler function
function appTimer()  
{    
	if (is_task && timer_enabled) // else: nop, idle, do nothing 
	{
		counter++;  
		if (counter >= timer)  {     // task time is limited
			theEnd();
			return;
		}	
		if (prg_type == '1')   {   // '1' => Time. display time on Progress Bar  
			let rest= timer-counter;   // this is remaining time 
			let percent= Math.round(100*rest/timer);    
			document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
			document.getElementById('nav_items').textContent= stringTime(rest);    
		}
	}  
}	


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  var hh = Math.floor(total/(60*60));         
  var div_min = total % (60 * 60);
  var mm = Math.floor(div_min / 60);       
  var div_sec = div_min % 60;
  var ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// raised when page is resized. see html: <body onresize="pageResize()"
// set task layout
function pageResize()  
{ 
	if (is_task)  { 
		let canvas= document.getElementById('canvas');
		canvas.width= window.innerWidth;  
		canvas.style.width= canvas.width + 'px';	     
		canvas.height= window.innerHeight - document.getElementById('footer').offsetHeight; // document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight  
		canvas.style.height= canvas.height + 'px';  // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 		
		
		drawStim();
	}
}	


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  if (is_storage && localStorage.getItem('eqlt-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('eqlt-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' EQLT results stored in your web browser.';
  }
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		       
  document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';	
	document.getElementById('ins_back').style.display= is_back ? 'block' : 'none';
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'item_cnt':
			if (e.validity.valid)  {
				msg.textContent= 'The number of stimuli in the task'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;  				
						
    case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'Timer is disabled';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;
		
    case 'timer':
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value); 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;  		
			   		
    case 'is_back':     		
			msg.textContent= e.checked ? 'Yes' : 'No. Returning to previous items is not allowed.';    
			break;
        
    case 'prg_type':                                                                                                                                            
      if (e.value == '0')  {
        msg.textContent= 'It will be displayed how many items are left to the end.';
      }
      else if (e.value == '1')  {
        msg.textContent= 'The remaining time until the end of the task will be displayed. Only when timer is enabled.';
      }
      else  {
        msg.textContent= ''; //'Nothing';
      }
      break;    
							
		case 'demo': 
       msg.textContent= e.checked ? 'On. The percentage error will be displayed.' : 'Off';  //The error size in percentages will be displayed 
       break; 
		
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity.valid object in all input type="number"
function optionsValid()  
{ 
  let res= document.getElementById('item_cnt').validity.valid && document.getElementById('timer').validity.valid; 	
  return res;              
}


// start task (research) 
function startTask(e)   
{       			
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
	document.getElementById('task_section').style.display= 'block'; 	  
	document.getElementById('canvas').style.display= 'block';			    
		
	document.getElementById('progress').style.display= 'block';   //document.getElementById('progress').setAttribute('style', 'width:100%');	 
	if ( document.getElementById('prg_type').value == '2' || ( document.getElementById('prg_type').value == '1' && timer_enabled == false ) ) {
		document.getElementById('progress_holder').style.display= 'none'; 
	}	
	else  
	{
		document.getElementById('progress_holder').style.display= 'block';     
		document.getElementById('progress').setAttribute('style', 'width:100%');   		
		if (document.getElementById('prg_type').value == '0') {
			document.getElementById('progress').className='progress-bar progress-bar-info progress-bar-striped';     // items/total
		}
		else if (document.getElementById('prg_type').value == '1')  {
			document.getElementById('progress').className= 'progress-bar progress-bar-danger progress-bar-striped';  // remaining time   
		} 	
		document.getElementById('progress').setAttribute('style', 'width:0');  // reset progress bar width
	}					
	document.getElementById('nav_items').textContent='';  // reset   	
 
  setPar();
 
	stimGen();
			
  is_task= true;    // finally "activate" timer  	
  		
	drawStim();    // show first item (stimulus)
	  	 
}

  
// set internal application research parameters (global vars)
function setPar()  
{  
	item= 0;  			   
	counter= 0;       // reset timer counter	
	stim= [];	 	
	dv= [];
	mis= []; 
	k1= [];
  for (let i= 0; i< item_cnt; i++)   
  { // i => item
		stim[i]= [];
    mis[i]= true; 		
		for (let j = 0; j < 2; j++)  {     
      stim[i][j]= [];		
    }		
  } 
}


// https://varun.ca/polar-coords/
// generate all stimuli for each item in range 0..item_cnt-1
function stimGen()	
{			 	
	let canvas= document.getElementById('canvas');  // set default task layout
	canvas.width= window.innerWidth;  
	canvas.style.width= canvas.width + 'px';	     
	canvas.height= window.innerHeight - document.getElementById('footer').offsetHeight; // document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight  
	canvas.style.height= canvas.height + 'px';  // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 			
	let w= canvas.width, h= canvas.height;	
	let is_hor = w >= h ? true : false; // fixed during whole task	
	let r = is_hor ? Math.min(w/4, h/2) : Math.min ( w/2, h/4 );  // polar radius in pixels
	r-= 20;  // offset from edge: 2*radius of circle
	let whole= 2*Math.PI, x0, y0, angle, prior= 100;
	
	for (let i = 0; i < item_cnt; i++)    // i => item  
	{		
		do  {				
		  angle= getRndFloat(0, whole); // 1. point => angle in radians			
		} while ( Math.abs(angle - prior) / whole  < 0.05  );  // +/-5% is min offset from previous angle	
		prior= angle; // previous angle
		for (let j= 0; j < 2; j++)
		{
			if ( j == 0 ) 
			{	// 1. line			
				if (is_hor) {
					x0 = w / 4;
					y0 = h / 2; 
				}
				else  {
					x0 = w / 2;
					y0 = h / 4; 				
				}				
			}
			else  
			{	// 2. line
				if (is_hor) {
					x0 = 3 * w / 4;
					y0 = h / 2; 
				}
				else  {
					x0 = w / 2;
					y0 = h * 3 / 4; 				
				}				  
			}
			
			if (j == 1 && !demo)  {
				angle+= Math.PI/2;   // orthogonal (right angle)
			}
			
			if (!demo && j == 1)   {  // decrease size of line 2 by half (100%) 
				stim[i][j][0]= x0;   // px 1. point: x   
				stim[i][j][1]= y0;   // px 1. point: y			    
				stim[i][j][2]= x0 + r * Math.cos(angle);   // px 2. point: x
				stim[i][j][3]= y0 + r * Math.sin(angle);   // px 2. point: y				
			}
			else  {
				stim[i][j][0]= x0 - r * Math.cos(angle);   // px 1. point: x  
				stim[i][j][1]= y0 - r * Math.sin(angle);   // px 1. point: y			    
				stim[i][j][2]= x0 + r * Math.cos(angle);   // px 2. point: x
				stim[i][j][3]= y0 + r * Math.sin(angle);   // px 2. point: y
			}
      			
			if ( j == 0 )  {
			  k1[i]= ( stim[i][0][3] - stim[i][0][1] ) / ( stim[i][0][2] - stim[i][0][0] );  // first fixed line: dy / dy
      }
						
			stim[i][j][0]= 100 * stim[i][j][0] / w;    // convert x pixels into units: percents of canvas width
	    stim[i][j][1]= 100 * stim[i][j][1] / h;    // convert y pixels into units: percents of canvas height
	    stim[i][j][2]= 100 * stim[i][j][2] / w;    // convert x pixels into units: percents of canvas width
	    stim[i][j][3]= 100 * stim[i][j][3] / h;    // convert y pixels into units: percents of canvas height
						     
		}

		dv[i]= demo ? 0 : null; 		
	}

}


// get random integer from min to max range
function getRndInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1)) + min;
}


/* https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Math/random
  Returns a float random number between the specified values. 
  The returned value is no lower than (and may possibly equal) min, and is less than (and not equal) max.
  Returns a random number between min (inclusive) and max (exclusive)  */
function getRndFloat(min, max) {
  return Math.random() * (max - min) + min;
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
	
	if (document.getElementById('item_cnt').validity.valid)  {  
	  item_cnt= parseInt(document.getElementById('item_cnt').value, 10);			
  } 
		
	timer_enabled= (document.getElementById('timer_enabled').checked) ? true: false;  // timer is enabled or disabled
	
	if (timer_enabled && document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }	
	
	is_back= document.getElementById('is_back').checked ? true: false;   // enable Go back (previous item)      
  	
	prg_type= document.getElementById('prg_type').value;   //  type of information displayed on Progress Bar	
		
  demo= document.getElementById('demo').checked ? true: false;	
	
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;

}


// goto previous item (stimulus)
function priorItem()  
{      
  if (item > 0)  {
    item--;  
    drawStim();     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{    		
	item++;					
	if (item < item_cnt)  {         		 
    drawStim();     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


// draw (repaint) current stimulus           
function drawStim() 
{    		
  const len = 10;     // mouse pointer radius range in px
	
	let canvas= document.getElementById('canvas'), ctx= canvas.getContext('2d'), w= canvas.width, h= canvas.height;	
	ctx.clearRect(0, 0, w, h);  //ctx.clearRect(0, 0, canvas.clientWidth, canvas.clientHeight);
  	
	ctx.lineWidth= 3;  	
	ctx.strokeStyle= '#000'; 
	let x1, y1, x2, y2, k2, ort, m=[]; // m= line middle coord
	for (let j= 0; j < 2; j++) 
	{
		x1 = stim[item][j][0] * w / 100;  // transform units: percent of canvas.width => pixels
		y1 = stim[item][j][1] * h / 100;	// transform units: percent of canvas.height => px			
		x2 = stim[item][j][2] * w / 100;  // transform units: percent of canvas.width => px
		y2 = stim[item][j][3] * h / 100;	// transform units: percent of canvas.height => px	 	 								  		
		ctx.beginPath();  	 
		ctx.moveTo(x1, y1);   // 1. point (line 1 or line2)  
		ctx.lineTo(x2, y2);   // draw line between 1. point and 2. point of line (line1 or line2)  	    	
		if ( j == 0 ) {
			ort = Math.atan(-1/k1[item]); 
		}
		else  {
			k2= ( stim[item][1][3] - stim[item][1][1] ) / ( stim[item][1][2] - stim[item][1][0] );   // second variable line: dy / dy
			ort	= Math.atan( -1 / k2 );   //  orthogonal (perpendicular) angle for 2. line
		}		
		ctx.moveTo(x1 - len * Math.cos(ort), y1 - len * Math.sin(ort));
		ctx.lineTo(x1 + len * Math.cos(ort), y1 + len * Math.sin(ort)); 		
		ctx.stroke();
		
		ctx.beginPath(); 	
		ctx.strokeStyle= j == 1 ? '#00f' : '#000'; // When you set the strokeStyle and/or fillStyle property, the new value becomes the default for all shapes being drawn from then on. For every shape you want in a different color, you will need to reassign the fillStyle or strokeStyle property.		
		ctx.moveTo(x2 - len * Math.cos(ort), y2 - len * Math.sin(ort));
		ctx.lineTo(x2 + len * Math.cos(ort), y2 + len * Math.sin(ort)); 			
		ctx.stroke();			
    if ( demo )  {
			m.push( divLine(x1, x2, -1) );  //  line middle(x): (x1 + x2 ) / 2   
		  m.push( divLine(y1, y2, -1) );	//  (y1 + y2) / 2 		
		}		
	}
	
	// update dependent variable
	let d1 = getDistance( stim[item][0][0] * w / 100, stim[item][0][1] * h / 100, stim[item][0][2] * w / 100, stim[item][0][3] * h / 100 );    // function getDistance(x1, y1, x2, y2)	
	let d2 = getDistance( x1, y1, x2, y2 );    // function getDistance(x1, y1, x2, y2)	
  let delta= d2 - d1;
	if ( delta >= 0 ) {  // line 2 >= line 1 
		dv[item]= 100 * ( Math.abs (delta) / d1); 	// dependent variable is error in percents. base unit is d1 length
	}
	else  {
		dv[item]= 100 * ( Math.abs (delta) / d2); 	// dependent variable is error in percents. base unit is d2 length	
	}
	if (demo)  {		// https://jsfiddle.net/dwahlin/GDTBy/		
    ctx.fillStyle= '#000';  //ctx.strokeStyle= '#000';
		ctx.font = '1em Arial';   
		ctx.textAlign = 'middle';
		ctx.textBaseline= 'bottom';
		ctx.save();
		
		ctx.translate(m[0],m[1]);
		ctx.rotate( Math.atan(k1[item]));
		ctx.fillText('Length= ' + d1.toFixed(2), 0, 0);
		ctx.rotate( Math.atan(-k1[item]) );		
		ctx.translate(-m[0],-m[1]);
	  
		ctx.translate(m[2],m[3]);		
		ctx.rotate(Math.atan(k2));		
		ctx.fillText('Length= ' + d2.toFixed(2) + ' Error= ' + dv[item].toFixed(2) + '%' , 0, 0);  //ctx.fillText('Error= ' + dv[item].toFixed(2), 0, 0);
		//ctx.rotate( Math.atan(-k2) );		
		//ctx.translate(-m[2],-m[3]);
		
		ctx.restore(); 
	} 

	updateFooter();
	
	/*if (demo)  {		// https://jsfiddle.net/dwahlin/GDTBy/		
    ctx.fillStyle= '#000';  //ctx.strokeStyle= '#000';
		ctx.font = '1em Arial';   		
		ctx.textAlign = 'middle';
		ctx.textBaseline= 'bottom';
			
		ctx.save();
		let xs= divLine(x1, x2, -1), ys= divLine(y1, y2, -1);
		ctx.translate(xs,ys);
		ctx.rotate(Math.atan(k2));	
		ctx.fillText('Error= ' + dv[item].toFixed(2), 0, 0);
		//ctx.translate(-xs,-ys);
		ctx.restore(); 
	}*/		
		  
}


/* divide line 
  first => first point (P1) 
  second => second point (P2)
  n => (distance(P1-Q)/ distance(Q-P2)) = lambda = -m/n = negative ratio length: P1-Q/Q-P2
  returns: coordinate of point (Q) between first and second point   */   
function divLine(first,second,n)  {     
  return (first-n*second)/(1-n);   // Math.round((first-n*second)/(1-n));
}


// update footer when show each stimulus in task. update buttons and update progress bar
function updateFooter()
{  
	document.getElementById('prior').style.display= item > 0 && is_back ? 'inline': 'none';  // set footer navigation
	let btn= document.getElementById('next');     
	if (item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update 'Next item' image on 'right arrow' button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Go forward';
	}						
  if (prg_type == '0' )  {    // update Progress bar => show current item count and remaining item count on Progress Bar (default)                                                                                                                                                       
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent= 'Item= ' + (item + 1) + ' Remaining= ' + (item_cnt - item-1);  
	}
}


// end of task (research)
function theEnd()  
{   

  is_task= false;    	
	
	document.getElementById('task_section').style.display= 'none';	
	document.getElementById('canvas').style.display= 'none'; 
	//document.getElementById('footer').style.display= 'none';
  document.getElementById('end').style.display= 'block';   
  document.getElementById('result_wrapper').style.display= 'block'; 
	
  let blank= 0, rows = [], err, arr= [];  
	// compute points (and total score) for each item in task.  
	for (let i = 0; i < item_cnt; i++)  
	{		
		if ( dv[i] == null )  {
      err= '';
		}
		else {
			err= dv[i].toFixed(2); 
			arr.push(dv[i]);
		}
		if (mis[i] == true)  {
      rows[i]= [ parseInt(i+1,10), err, 'No' ];  // => [ 'Item', 'Error (%)' , 'Moved' ]  
			blank++;			
		}
		else {	
			rows[i]= [ parseInt(i+1,10), err, 'Yes' ];  // => [ 'Item', 'Error (%)' , 'Moved' ]      
		}		
	}		

	createTableArray(document.getElementById('table1'), rows,  ['Item', 'Error (%)', 'Moved']); 
	 		
	document.getElementById('table1').style.display= 'block';
		
	// calculate decriptive statistics for the errors (%) => table2 + chart1 
	rows = [];
	let n= arr.length;
	let m = (n > 0) ? mean(arr).toFixed(2) : '';     // mean only for "valid" items in task
	let sd = (n > 1) ? stDev(arr, m).toFixed(2): ''; // standard deviation only for valid items in task  	
	
	let tmp= ['Variable', 'Average', 'Sample Standard Deviation', 'N']; // header
	rows.push( ['Error (%)', m, sd, n] );  // last row
	
	createTableArray(document.getElementById('table2'), rows, tmp);   
	
  // "chart" => repaint "chart" 
  let ok= (100-m).toFixed(2);	
	document.getElementById('chart_2').setAttribute('style', 'width:' + m + '%');
	document.getElementById('chart_1').setAttribute('style', 'width:' + ok + '%');    // !important
	document.getElementById('chart').title= 'The average error is= ' + m + '%';  	
		
	// update 'score' label => above "Repeat" button
	document.getElementById('score').textContent= 'Average error= ' + m + '% Sample Standard Deviation= ' + sd + ' Not-moved items= ' + blank; 
	   					
	try 
  {
    if (store) {
			setStorage(m, sd, blank); 
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }    	 
  
}


// return arithmetic mean (mean, average) of array (arr) 
function mean(arr)  
{	
	let sum=0;
	for (let i in arr)  {
		sum+= parseFloat(arr[i]);
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation
// arr = array argument; mean = average (mean)
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
	  sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 m  => The average error (%) 
   sd => Sample Standard Deviation
   not_moved => number of not-moved items  */ 
function setStorage(m, sd, not_moved)   
{    
	let oldItems = JSON.parse(localStorage.getItem('eqlt-v1.0')) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
      'error-m' : m, 	
			'error-sd' : sd,
      'not-moved': not_moved,           
			'demo' : demo  ? 1: 0,    
    	'items' : item_cnt,  			     
			'timer-time' : timer_enabled ? document.getElementById('timer').value : '',     
		 	'is-back' : is_back ? 1: 0			
  };

  oldItems.push(newItem);

  localStorage.setItem('eqlt-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);   
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement("table");  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement("th");      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('eqlt-v1.0')  ?  JSON.parse(localStorage.getItem('eqlt-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no EQLT data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' EQLT results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved EQLT results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('eqlt-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data  ';
 }  
}


// return distance from point P1(x1,y1) to point P2(x2,y2)
function getDistance(x1, y1, x2, y2)  {
  return Math.sqrt( Math.pow(x1-x2, 2) + Math.pow(y1-y2, 2) ); // total line distance (px) 
}	

 
/* when user move blue end of line 2. line 2 have variable length and have blue end region
   canvas => mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
   this event is fired by html canvas element => document.getElementById('canvas') 
	 <canvas id="canvas" onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>  */
function mouseMove(e)
{     				
	if (!is_move) { return; } 
	let xm= e.offsetX, ym= e.offsetY;   // mouse pointers. Note: e.offsetX == e.clientX;  e.offsetY == e.clientY 	
	let canvas= document.getElementById('canvas'), w= canvas.width, h= canvas.height;	 
	
	/* let d = getDistance( xm, ym, stim[item][1][2] * w / 100, stim[item][1][3] * h / 100 );    // function getDistance(x1, y1, x2, y2)	
	is_move= d <= 20 ? true: false;
	if ( is_move )   // https://stackoverflow.com/questions/36391187/rotate-line-with-fixed-size-by-mouse
	{	*/ 	 
	
	let x0= stim[item][1][2] * w /100;    // start x (px) line 2
	let	y0= stim[item][1][3] * h / 100;   // start y (px)	line 2			
	let dx = xm - x0,  dy = ym - y0;  // Just determine the angle between the start point of the line and mouse position:
	let angle = Math.atan2(dy, dx);		
	let r = getDistance(xm, ym, x0, y0);  // function getDistance(x1, y1, x2, y2)
	let x = x0 + r * Math.cos(angle), y = y0 + r * Math.sin(angle);  // Then render the line using r pixel radius:
	stim[item][1][2]= 100 * x / w; // transform px into units canvas.width  %		
	stim[item][1][3]= 100 * y / h; // transform px into units canvas.height % 
		
	mis[item]= false;
	drawStim();  // repaint stimulus				
  
	/* }	
	else  {
	  canvas.style.backgroundColor= '#fff';	
    document.getElementById('canvas').style.cursor= 'auto'; //document.getElementById('canvas').style.cursor= 'initial';  // because of IE11		
	} */ 	
}


// html <canvas id="canvas" onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>                        
function mouseDown(e)  
{	 
	let canvas= document.getElementById('canvas'), w= canvas.width, h= canvas.height;
	let x = stim[item][1][2] * w / 100;  // transform units: percent of canvas.width => x pixels
	let y = stim[item][1][3] * h / 100;	 // transform units: percent of canvas.height => y pixels	
		// e.offsetX and e.offsetY => mouse positions x, y	
	let d= getDistance(e.offsetX, e.offsetY, x, y); // distance from Point (mouse positions) and Point (line2 blue end) in pixels
	is_move= d <= 10 ? true: false;
	if (is_move)  {
		canvas.style.cursor= 'move';  //  'ew-resize'  or  'ns-resize'  
		canvas.style.backgroundColor= '#e5e5e5';	
	}
	else  {
	  canvas.style.backgroundColor= '#fff';	
    document.getElementById('canvas').style.cursor= 'auto'; //document.getElementById('canvas').style.cursor= 'initial';  // because of IE11		
	}	
}


// html <canvas id="canvas" onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>                        
function mouseUp()  {
	is_move= false;	
	document.getElementById('canvas').style.backgroundColor= '#fff';
  document.getElementById('canvas').style.cursor= 'auto';  //document.getElementById('canvas').style.cursor= 'initial';  // because of IE11 
  drawStim();
}
